"""Environment configuration variables that can be passed when executing a lane."""
import os

INTERNAL_LOGGER_NAME = 'SPARKLANES'
"""The logger's name under which internal events will be logged."""

SPARK_APP_NAME = 'sparklanes.app'
"""The app name using which the default SparkContext/SparkSession will be instantiated"""

UNNAMED_LANE_NAME = 'UnnamedLane'
"""Default name of a lane, if no custom name is specified."""

UNNAMED_BRANCH_NAME = 'UnnamedBranch'
"""Default name of a branch, if no custom branch is specified"""

VERBOSE_TESTING = False
"""Specifies if output should be printed to console when running tests"""

INIT_SPARK_ON_IMPORT = True
"""Specifies if a default SparkContext/SparkSession should be instantiated upon import sparklanes"""

for v in locals().copy():
    if not v.startswith('__') and os.getenv(v):
        env_var = os.getenv(v)
        if env_var:
            var_type = type(locals()[v])
            try:
                if var_type == bool:
                    locals()[v] = bool(int(env_var))
                else:
                    locals()[v] = type(var_type)(env_var)
            except (ValueError, TypeError) as exc:
                raise ValueError('Invalid value for environment variable `%s`. Must be `%s`.'
                                 % (v, str(var_type) if not var_type == bool else '0/1'))



"""Lane and Branch classes. TODO: Better logging"""
from importlib import import_module
from inspect import isclass

import yaml
from schema import SchemaError

from six import string_types

from sparklanes._framework.errors import LaneSchemaError, LaneImportError
from sparklanes._framework.validation import validate_schema

from .env import INTERNAL_LOGGER_NAME
from .errors import LaneExecutionError
from .log import make_default_logger
from .task import LaneTask, LaneTaskThread
from .validation import validate_params


class Lane(object):
    """Used to build and run data processing lanes (i.e. pipelines).
    Public methods are chainable."""

    def __init__(self, name='UnnamedLane', run_parallel=False):
        """
        Parameters
        ----------
        name : str
            Custom name of the lane
        run_parallel : bool
            Indicates, whether the tasks in a Lane shall be executed in parallel.
            Does not affect branches inside the lane (`run_parallel` must be indicated in the
            branches themselves)
        """
        if not isinstance(name, string_types):
            raise TypeError('`name` must be a string')
        self.name = name
        self.run_parallel = run_parallel

        self.tasks = []

    def __str__(self):
        """Generates a readable string using the tasks/branches inside the lane, i.e. builds a
        string the showing the tasks and branches in a lane"""
        task_str = '=' * 80 + '\n'

        def generate_str(lane_or_branch, prefix='\t', out=''):
            """Recursive string generation"""
            out += prefix + lane_or_branch.name
            if lane_or_branch.run_parallel:
                out += ' (parallel)'
            out += '\n'

            for task in lane_or_branch.tasks:
                if isinstance(task, Branch):
                    out += generate_str(task, prefix + prefix[0])
                elif isinstance(task['cls_or_branch'], Branch):
                    out += generate_str(task['cls_or_branch'], prefix + prefix[0])
                else:
                    out += prefix + ' >' + task['cls_or_branch'].__name__ + '\n'

            return out

        task_str += generate_str(self) + '=' * 80

        return task_str

    def __validate_task(self, cls, entry_mtd_name, args, kwargs):
        """Checks if a class is a task, i.e. if it has been decorated with `sparklanes.Task`, and if
        the supplied args/kwargs match the signature of the task's entry method.

        Parameters
        ----------
        cls : LaneTask
        entry_mtd_name : str
            Name of the method, which is called when the task is run
        args : list
        kwargs : dict
        """
        if not isclass(cls) or not issubclass(cls, LaneTask):
            raise TypeError('Tried to add non-Task `%s` to a Lane. Are you sure the task was '
                            'decorated with `sparklanes.Task`?' % str(cls))

        validate_params(cls, entry_mtd_name, *args, **kwargs)

    def add(self, cls_or_branch, *args, **kwargs):
        """Adds a task or branch to the lane.

        Parameters
        ----------
        cls_or_branch : Class
        *args
            Variable length argument list to be passed to `cls_or_branch` during instantiation
        **kwargs
            Variable length keyword arguments to be passed to `cls_or_branch` during instantiation

        Returns
        -------
        self: Returns `self` to allow method chaining
        """
        if isinstance(cls_or_branch, Branch):
            self.tasks.append(cls_or_branch)  # Add branch with already validated tasks
        else:
            # Validate
            self.__validate_task(cls_or_branch, '__init__', args, kwargs)
            # Append
            self.tasks.append({'cls_or_branch': cls_or_branch, 'args': args, 'kwargs': kwargs})

        return self

    def run(self):
        """Executes the tasks in the lane in the order in which they have been added, unless
        `self.run_parallel` is True, then a thread is spawned for each task and executed in
        parallel (note that task threads are still spawned in the order in which they were added).
        """
        logger = make_default_logger(INTERNAL_LOGGER_NAME)
        logger.info('\n%s\nExecuting `%s`\n%s\n', '-'*80, self.name, '-'*80)
        logger.info('\n%s', str(self))

        threads = []

        if not self.tasks:
            raise LaneExecutionError('No tasks to execute!')

        for task_def_or_branch in self.tasks:
            if isinstance(task_def_or_branch, Branch):
                task_def_or_branch.run()
            elif isinstance(task_def_or_branch['cls_or_branch'], Branch):  # Nested Branch
                task_def_or_branch['cls_or_branch'].run()
            else:
                task = task_def_or_branch['cls_or_branch'](*task_def_or_branch['args'],
                                                           **task_def_or_branch['kwargs'])
                if self.run_parallel:
                    threads.append(LaneTaskThread(task))
                else:
                    task()

        if threads:
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()

        logger.info('\n%s\nFinished executing `%s`\n%s', '-'*80, self.name, '-'*80)

        return self


class Branch(Lane, object):
    """Branches can be used to split task lanes into branches, which is e.g. useful if part of the
    data processing pipeline should be executed in parallel, while other parts should be run in
    subsequent order."""

    def __init__(self, name='UnnamedBranch', run_parallel=False):
        """
        Parameters
        ----------
        name (str): Custom name of the branch
        run_parallel (bool): Indicates if the task in the branch shall be executed in parallel
        args (List[object])
        """
        super(Branch, self).__init__(name=name, run_parallel=run_parallel)


def build_lane_from_yaml(path):
    """Builds a `sparklanes.Lane` object from a YAML definition file.

    Parameters
    ----------
    path: str
        Path to the YAML definition file

    Returns
    -------
    Lane
        Lane, built according to definition in YAML file
    """
    # Open
    with open(path, 'rb') as yaml_definition:
        definition = yaml.load(yaml_definition)

    # Validate schema
    try:
        validate_schema(definition)
    except SchemaError as exc:
        raise LaneSchemaError(**exc.__dict__)

    def build(lb_def, branch=False):
        """Function to recursively build the `sparklanes.Lane` object from a YAML definition"""
        init_kwargs = {k: lb_def[k] for k in (a for a in ('run_parallel', 'name') if a in lb_def)}
        lane_or_branch = Lane(**init_kwargs) if not branch else Branch(**init_kwargs)

        for task in lb_def['tasks']:
            if 'branch' in task:
                branch_def = task['branch']
                lane_or_branch.add(build(branch_def, True))
            else:
                sep = task['class'].rfind('.')
                if sep == -1:
                    raise LaneImportError('Class must include its parent module')
                mdl = task['class'][:sep]
                cls_ = task['class'][sep + 1:]

                try:
                    cls = getattr(import_module(mdl), cls_)
                except ImportError:
                    raise LaneImportError('Could not find module %s' % mdl)
                except AttributeError:
                    raise LaneImportError('Could not find class %s' % cls_)

                args = task['args'] if 'args' in task else []
                args = [args] if not isinstance(args, list) else args
                kwargs = task['kwargs'] if 'kwargs' in task else {}
                lane_or_branch.add(cls, *args, **kwargs)

        return lane_or_branch

    return build(definition['lane'])



"""Used to allow sharing of SparkContext and SparkSession, to avoid having to "getOrCreate" them
again and again for each task. This way, they can just be imported and used right away."""
# pylint: disable=invalid-name,too-many-arguments
from pyspark import SparkContext, PickleSerializer, BasicProfiler
from pyspark.sql import SparkSession
from six import PY2

from sparklanes._framework.env import INIT_SPARK_ON_IMPORT, SPARK_APP_NAME


class SparkContextAndSessionContainer(object):
    """Container class holding SparkContext and SparkSession instances, so that any changes
    will be propagated across the application"""
    sc = None
    spark = None

    def __new__(cls, *args, **kwargs):
        if cls is SparkContextAndSessionContainer:
            raise TypeError('SparkSession & SparkContext container class may not be instantiated.')

        return object.__new__(cls, *args, **kwargs) if PY2 else object.__new__(cls)

    @classmethod
    def set_sc(cls, master=None, appName=None, sparkHome=None, pyFiles=None, environment=None,
               batchSize=0, serializer=PickleSerializer(), conf=None, gateway=None, jsc=None,
               profiler_cls=BasicProfiler):
        """Creates and initializes a new `SparkContext` (the old one will be stopped).
        Argument signature is copied from `pyspark.SparkContext
        <https://spark.apache.org/docs/latest/api/python/pyspark.html#pyspark.SparkContext>`_.
        """
        if cls.sc is not None:
            cls.sc.stop()
        cls.sc = SparkContext(master, appName, sparkHome, pyFiles, environment, batchSize,
                              serializer,
                              conf, gateway, jsc, profiler_cls)
        cls.__init_spark()

    @classmethod
    def set_spark(cls, master=None, appName=None, conf=None, hive_support=False):
        """Creates and initializes a new `SparkSession`. Argument signature is copied from
        `pyspark.sql.SparkSession
        <https://spark.apache.org/docs/latest/api/python/pyspark.sql.html#pyspark.sql.SparkSession>`_.
        """
        sess = SparkSession.builder
        if master:
            sess.master(master)
        if appName:
            sess.appName(appName)
        if conf:
            sess.config(conf=conf)
        if hive_support:
            sess.enableHiveSupport()

        cls.spark = sess.getOrCreate()

    @classmethod
    def init_default(cls):
        """Create and initialize a default SparkContext and SparkSession"""
        cls.__init_sc()
        cls.__init_spark()

    @classmethod
    def __init_sc(cls):
        cls.sc = SparkContext(appName=SPARK_APP_NAME).getOrCreate()

    @classmethod
    def __init_spark(cls):
        cls.spark = SparkSession.builder.appName(SPARK_APP_NAME).getOrCreate()


if INIT_SPARK_ON_IMPORT:
    SparkContextAndSessionContainer.init_default()



"""A simple 'main' file, that will be submitted to spark. It will execute the lane as defined in
the YAML lane definition file."""
# pylint: disable=missing-docstring
from argparse import ArgumentParser

from sparklanes import build_lane_from_yaml


def main():
    args = parse_args()
    build_lane_from_yaml(args['lane']).run()


def parse_args():
    parser = ArgumentParser()
    parser.add_argument('-l', '--lane',
                        help='Relative or absolute path to the lane definition YAML file',
                        type=str,
                        required=True)

    return parser.parse_args().__dict__


if __name__ == '__main__':
    main()



"""Module handling logging. TODO: improve logging. Allow configuration, etc."""
import logging
import sys

from .env import INTERNAL_LOGGER_NAME


def make_default_logger(name=INTERNAL_LOGGER_NAME, level=logging.INFO,
                        fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s'):
    """Create a logger with the default configuration"""
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if not logger.handlers:
        handler = logging.StreamHandler(sys.stderr)
        handler.setLevel(level)
        formatter = logging.Formatter(fmt)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger



"""Module that allows submitting lanes to spark using YAML definitions"""
import argparse
import logging
import os
import re
import shutil
import sys
import tempfile
from subprocess import call, STDOUT

SPARK_SUBMIT_FLAGS = ['verbose', 'supervised']


def submit_to_spark():
    """Console-script entry point"""
    _package_and_submit(sys.argv[1:])


def _package_and_submit(args):
    """
    Packages and submits a job, which is defined in a YAML file, to Spark.

    Parameters
    ----------
    args (List): Command-line arguments
    """
    args = _parse_and_validate_args(args)

    logging.debug(args)
    dist = __make_tmp_dir()
    try:
        __package_dependencies(dist_dir=dist, additional_reqs=args['requirements'],
                               silent=args['silent'])
        __package_app(tasks_pkg=args['package'],
                      dist_dir=dist,
                      custom_main=args['main'],
                      extra_data=args['extra_data'])
        __run_spark_submit(lane_yaml=args['yaml'], dist_dir=dist, spark_args=args['spark_args'],
                           silent=args['silent'])

    except Exception as exc:
        __clean_up(dist)
        raise exc
    __clean_up(dist)


def _parse_and_validate_args(args):
    """
    Parse and validate arguments. During validation, it is checked whether the given
    files/directories exist, while also converting relative paths to absolute ones.

    Parameters
    ----------
    args (List): Command-line arguments
    """
    parser = argparse.ArgumentParser(description='Submitting a lane to spark.')
    parser.add_argument('-y', '--yaml', type=str, required=True,
                        help='Path to the yaml definition file.')
    parser.add_argument('-p', '--package', type=str, required=True,
                        help='Path to the python package containing your tasks.')
    parser.add_argument('-r', '--requirements', type=str, required=False,
                        help='Path to a `requirements.txt` specifying any additional dependencies '
                             'of your tasks.')
    parser.add_argument('-e', '--extra-data', nargs='*', required=False,
                        help='Path to any additional files or directories that should be packaged '
                             'and sent to Spark.')
    parser.add_argument('-m', '--main', type=str, required=False,
                        help='Path to a custom main python file')
    parser.add_argument('-s', '--spark-args', nargs='*', required=False,
                        help='Any additional arguments that should be sent to Spark via '
                             'spark-submit. '
                             '(e.g. `--spark-args executor-memory=20G total-executor-cores=100`)')
    parser.add_argument('--silent', help='If set, no output will be sent to console',
                        action='store_true')
    args = parser.parse_args(args).__dict__

    # Check/fix files/dirs
    args['package'] = __validate_and_fix_path(args['package'], check_dir=True)
    for param in ('yaml', 'requirements', 'main'):
        args[param] = __validate_and_fix_path(args[param], check_file=True)
    if args['extra_data']:
        for i in range(len(args['extra_data'])):
            args['extra_data'][i] = __validate_and_fix_path(args['extra_data'][i],
                                                            check_file=True, check_dir=True)

    # Check if python package
    if not os.path.isfile(os.path.join(args['package'], '__init__.py')):
        raise SystemExit('Could not confirm `%s` is a python package. Make sure it contains an '
                         '`__init__.py`.')

    # Check/fix spark args
    if args['spark_args']:
        args['spark_args'] = __validate_and_fix_spark_args(args['spark_args'])

    return args


def __validate_and_fix_path(path, check_file=False, check_dir=False):
    """Check if a file/directory exists and converts relative paths to absolute ones"""
    # pylint: disable=superfluous-parens
    if path is None:
        return path
    else:
        abs_path = path
        if not os.path.isabs(abs_path):
            abs_path = os.path.abspath(os.path.join(os.path.abspath(os.curdir), abs_path))
        if not (os.path.isfile(abs_path) if check_file else False) \
                and not (os.path.isdir(abs_path) if check_dir else False):
            raise SystemExit('Path `%s` does not exist' % path)

    return abs_path


def __validate_and_fix_spark_args(spark_args):
    """
    Prepares spark arguments. In the command-line script, they are passed as for example
    `-s master=local[4] deploy-mode=client verbose`, which would be passed to spark-submit as
    `--master local[4] --deploy-mode client --verbose`

    Parameters
    ----------
    spark_args (List): List of spark arguments

    Returns
    -------
    fixed_args (List): List of fixed and validated spark arguments
    """
    pattern = re.compile(r'[\w\-_]+=.+')
    fixed_args = []
    for arg in spark_args:
        if arg not in SPARK_SUBMIT_FLAGS:
            if not pattern.match(arg):
                raise SystemExit('Spark argument `%s` does not seem to be in the correct format '
                                 '`ARG_NAME=ARG_VAL`, and is also not recognized to be one of the'
                                 'valid spark-submit flags (%s).' % (arg, str(SPARK_SUBMIT_FLAGS)))
            eq_pos = arg.find('=')
            fixed_args.append('--' + arg[:eq_pos])
            fixed_args.append(arg[eq_pos + 1:])
        else:
            fixed_args.append('--' + arg)

    return fixed_args


def __make_tmp_dir():
    """
    Create a temporary directory where the packaged files will be located

    Returns
    -------
    tmp_dir (str): Absolute path to temporary directory
    """
    tmp_dir = tempfile.mkdtemp()
    logging.debug('Created temporary dir: `%s`', tmp_dir)

    return tmp_dir


def __package_dependencies(dist_dir, additional_reqs, silent):
    """
    Installs the app's dependencies from pip and packages them (as zip), to be submitted to spark.

    Parameters
    ----------
    dist_dir (str): Path to directory where the packaged libs shall be located
    additional_reqs (str): Path to a requirements.txt, containing any of the app's additional
        requirements
    silent (bool): Flag indicating whether pip output should be printed to console
    """
    logging.info('Packaging dependencies')
    libs_dir = os.path.join(dist_dir, 'libs')
    if not os.path.isdir(libs_dir):
        os.mkdir(libs_dir)

    # Get requirements
    req_txt = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'requirements-submit.txt')
    with open(req_txt, 'r') as req:
        requirements = req.read().splitlines()
    if additional_reqs:
        with open(additional_reqs, 'r') as req:
            for row in req:
                requirements.append(row)

    # Remove duplicates
    requirements = list(set(requirements))

    # Install
    devnull = open(os.devnull, 'w')
    outp = {'stderr': STDOUT, 'stdout': devnull} if silent else {}
    for pkg in requirements:
        cmd = ['pip', 'install', pkg, '-t', libs_dir]
        logging.debug('Calling `%s`', str(cmd))
        call(cmd, **outp)
    devnull.close()

    # Package
    shutil.make_archive(os.path.join(dist_dir, 'libs'), 'zip', libs_dir, './')


def __package_app(tasks_pkg, dist_dir, custom_main=None, extra_data=None):
    """
    Packages the `tasks_pkg` (as zip) to `dist_dir`. Also copies the 'main' python file to
    `dist_dir`, to be submitted to spark. Same for `extra_data`.

    Parameters
    ----------
    tasks_pkg (str): Path to the python package containing tasks
    dist_dir (str): Path to the directory where the packaged code should be stored
    custom_main (str): Path to a custom 'main' python file.
    extra_data (List[str]): List containing paths to files/directories that should also be packaged
        and submitted to spark
    """
    logging.info('Packaging application')

    # Package tasks
    tasks_dir_splits = os.path.split(os.path.realpath(tasks_pkg))
    shutil.make_archive(os.path.join(dist_dir, 'tasks'),
                        'zip',
                        tasks_dir_splits[0],
                        tasks_dir_splits[1])

    # Package main.py
    if custom_main is None:
        from . import _main
        main_path = _main.__file__
        if main_path[-3:] == 'pyc':
            main_path = main_path[:-1]
        shutil.copy(os.path.realpath(main_path),
                    os.path.join(dist_dir, 'main.py'))
    else:
        shutil.copy(os.path.realpath(custom_main),
                    os.path.join(dist_dir, 'main.py'))

    # Package _framework
    shutil.make_archive(os.path.join(dist_dir, '_framework'),
                        'zip',
                        os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..'),
                        './sparklanes/')

    # Package extra data
    if extra_data:
        for dat in extra_data:
            real_path = os.path.realpath(dat)
            target = os.path.join(dist_dir, os.path.split(real_path)[1])
            if os.path.isfile(real_path):
                shutil.copy(real_path, target)
            elif os.path.isdir(real_path):
                shutil.copytree(real_path, target)
            else:
                raise IOError('File `%s` not found at `%s`.' % (dat, real_path))


def __run_spark_submit(lane_yaml, dist_dir, spark_args, silent):
    """
    Submits the packaged application to spark using a `spark-submit` subprocess

    Parameters
    ----------
    lane_yaml (str): Path to the YAML lane definition file
    dist_dir (str): Path to the directory where the packaged code is located
    spark_args (str): String of any additional spark config args to be passed when submitting
    silent (bool): Flag indicating whether job output should be printed to console
    """
    cmd = ['spark-submit']

    # Supplied spark arguments
    if spark_args:
        cmd += spark_args

    # Packaged App & lane
    cmd += ['--py-files', 'libs.zip,_framework.zip,tasks.zip', 'main.py']
    if lane_yaml:
        cmd += ['--lane', lane_yaml]

    logging.info('Submitting to Spark')
    logging.debug(str(cmd))

    # Submit
    devnull = open(os.devnull, 'w')
    outp = {'stderr': STDOUT, 'stdout': devnull} if silent else {}
    call(cmd, cwd=dist_dir, **outp)
    devnull.close()


def __clean_up(dist_dir):
    """Delete packaged app"""
    shutil.rmtree(dist_dir)



"""Includes the `Task` decorator, the parent class `LaneTask` from which all tasks will inherit,
as well as the `_TaskCache`, which is used to share attributes between Task objects."""
import sys
from datetime import timedelta
from functools import WRAPPER_ASSIGNMENTS
from inspect import isclass
from threading import Thread, Lock
from time import time
from types import MethodType

from six import string_types, PY2, PY3

from .env import INTERNAL_LOGGER_NAME
from .errors import CacheError, TaskInitializationError, LaneExecutionError
from .log import make_default_logger


def Task(entry):  # pylint: disable=invalid-name
    """
    Decorator with which classes, who act as tasks in a `Lane`, must be decorated. When a class is
    being decorated, it becomes a child of `LaneTask`.

    Parameters
    ----------
    entry: The name of the task's "main" method, i.e. the method which is executed when task is run

    Returns
    -------
    wrapper (function): The actual decorator function
    """
    if not isinstance(entry, string_types):
        # In the event that no argument is supplied to the decorator, python passes the decorated
        # class itself as an argument. That way, we can detect if no argument (or an argument of
        # invalid type) was supplied. This allows passing of `entry` as both a named kwarg, and
        # as an arg. Isn't neat, but for now it suffices.
        raise TypeError('When decorating a class with `Task`, a single string argument must be '
                        'supplied, which specifies the "main" task method, i.e. the class\'s entry '
                        'point to the task.')
    else:
        def wrapper(cls):
            """The actual decorator function"""
            if isclass(cls):
                if not hasattr(cls, entry):  # Check if cls has the specified entry method
                    raise TypeError('Method `%s` not found in class `%s`.' % (entry, cls.__name__))

                # We will have to inspect the task class's `__init__` method later (by inspecting
                # the arg signature, before it is instantiated). In various circumstances, classes
                # will not have an unbound `__init__` method. Let's deal with that now already, by
                # assigning an empty, unbound `__init__` method manually, in order to prevent
                # errors later on during method inspection (not an issue in Python 3):
                # - Whenever a class is not defined as a new-style class in Python 2.7, i.e. a
                # sub-class of object, and it does not have a `__init__` method definition, the
                # class will not have an attribute `__init__`
                # - If a class misses a `__init__` method definition, but is defined as a
                # new-style class, attribute `__init__` will be of type `slot wrapper`, which
                # cannot be inspected (and it also doesn't seem possible to check if a method is of
                # type `slot wrapper`, which is why we manually define one).
                if not hasattr(cls, '__init__') or cls.__init__ == object.__init__:
                    init = MethodType(lambda self: None, None, cls) \
                        if PY2 else MethodType(lambda self: None, cls)
                    setattr(cls, '__init__', init)

                # Check for attributes that will be overwritten, in order to warn the user
                reserved_attributes = ('__getattr__', '__call__', '_entry_mtd', 'cache', 'uncache',
                                       'clear_cache', '_log_lock')
                for attr in dir(cls):
                    if attr in reserved_attributes:
                        make_default_logger(INTERNAL_LOGGER_NAME).warning(
                            'Attribute `%s` of class `%s` will be overwritten when decorated with '
                            '`sparklanes.Task`! Avoid assigning any of the following attributes '
                            '`%s`', attr, cls.__name__, str(reserved_attributes)
                        )

                assignments = {'_entry_mtd': entry,
                               '__getattr__': lambda self, name: TaskCache.get(name),
                               '__init__': cls.__init__,
                               '_log_lock': Lock()}
                for attr in WRAPPER_ASSIGNMENTS:
                    try:
                        assignments[attr] = getattr(cls, attr)
                    except AttributeError:
                        pass

                # Build task as a subclass of LaneTask
                return type('Task_%s' % cls.__name__, (LaneTask, cls, object), assignments)
            else:
                raise TypeError('Only classes can be decorated with `Task`')

        return wrapper


class LaneTask(object):
    """The super class of each task, from which all tasks inherit when being decorated with
    `sparklanes.Task`"""
    # pylint: disable=no-member
    def __new__(cls, *args, **kwargs):
        """Used to make sure the class will not be instantiated on its own. Instances of LaneTask
        should only exist as parents."""
        if cls is LaneTask:
            raise TaskInitializationError(
                "Task base `LaneTask` may not be instantiated on its own.")

        return object.__new__(cls, *args, **kwargs) if PY2 else object.__new__(cls)

    def __call__(self):
        """Used to make each task object callable, in order to execute tasks in a consistent
        manner. Calls the task's entry method and provides some logging."""
        logger = make_default_logger(INTERNAL_LOGGER_NAME)

        task_name = self.__name__ + '.' + self._entry_mtd
        logger.info('\n%s\nExecuting task `%s`\n%s',
                    '-'*80, task_name, '-'*80)
        start = time()
        res = getattr(self, self._entry_mtd)()
        passed = str(timedelta(seconds=(time() - start)))
        logger.info('\n%s\nFinished executing task `%s`. Execution time: %s\n%s',
                    '-'*80, task_name, passed, '-'*80)

        return res

    def cache(self, name, val, overwrite=True):
        """Assigns an attribute reference to all subsequent tasks. For example, if a task caches a
        DataFrame `df` using `self.cache('some_df', df)`, all tasks that follow can access the
        DataFrame using `self.some_df`. Note that manually assigned attributes that share the same
        name have precedence over cached attributes.

        Parameters
        ----------
        name : str
            Name of the attribute
        val
            Attribute value
        overwrite : bool
            Indicates if the attribute shall be overwritten, or not (if `False`, and
            a cached attribute with the given name already exists, `sparklanes.errors.CacheError`
            will be thrown).
        """
        if name in TaskCache.cached and not overwrite:
            raise CacheError('Object with name `%s` already in cache.' % name)
        TaskCache.cached[name] = val

    def uncache(self, name):
        """Removes an attribute from the cache, i.e. it will be deleted and becomes unavailable for
        all subsequent tasks.

        Parameters
        ----------
        name : str
            Name of the cached attribute, which shall be deleted
        """
        try:
            del TaskCache.cached[name]
        except KeyError:
            raise CacheError('Attribute `%s` not found in cache.' % name)

    def clear_cache(self):
        """Clears the entire cache"""
        TaskCache.cached = {}


class LaneTaskThread(Thread):
    """Used to spawn tasks as threads to be run in parallel."""

    def __init__(self, task):
        Thread.__init__(self)
        self.task = task
        self.exc = None
        self.daemon = True

    def run(self):
        """Overwrites `threading.Thread.run`, to allow handling of exceptions thrown by threads
        from within the main app."""
        self.exc = None
        try:
            self.task()
        except BaseException:
            self.exc = sys.exc_info()

    def join(self, timeout=None):
        """Overwrites `threading.Thread.join`, to allow handling of exceptions thrown by threads
        from within the main app."""
        Thread.join(self, timeout=timeout)
        if self.exc:
            msg = "Thread '%s' threw an exception `%s`: %s" \
                  % (self.getName(), self.exc[0].__name__, self.exc[1])
            new_exc = LaneExecutionError(msg)

            if PY3:
                raise new_exc.with_traceback(self.exc[2])  # pylint: disable=no-member
            else:
                raise (new_exc.__class__, new_exc, self.exc[2])  # pylint: disable=raising-bad-type


class TaskCache(object):
    """Serves as the attribute cache of tasks, which is accessed using the tasks'
    `__getattr__` method."""
    cached = {}

    def __new__(cls, *args, **kwargs):
        """Used to make sure that TaskCache will not be instantiated."""
        if cls is TaskCache:
            raise CacheError("`TaskCache` may not be instantiated and only provides static access.")

        return object.__new__(cls, *args, **kwargs) if PY2 else object.__new__(cls)

    @staticmethod
    def get(name):
        """Retrieves an object from the cache.

        Parameters
        ----------
        name : str
            Name of the object to be retrieved

        Returns
        -------
        object
        """
        try:
            return TaskCache.cached[name]
        except KeyError:
            raise AttributeError('Attribute `%s` not found in cache or object.' % name)
